"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.diff = diff;
exports.getColor = getColor;
exports.language = void 0;
exports.parse = parse;
exports.ready = ready;
exports.toString = toString;

var _starryNight = require("@wooorm/starry-night");

var _darkStyle = _interopRequireDefault(require("./dark-style"));

var _wcwidth = _interopRequireDefault(require("wcwidth"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const starryNight = (0, _starryNight.createStarryNight)(_starryNight.all);
let starryNightCache = null;

async function ready() {
  starryNightCache = await starryNight;
}

const handler = {
  get(_, language) {
    return (code, ...rest) => {
      return {
        language,
        spans: Array.from(code),
        nodes: rest
      };
    };
  }

};
const language = new Proxy({}, handler);
exports.language = language;

function parse(code, options) {
  const raw = integrate(reindent(code));
  if (starryNightCache == null) throw new Error(`you must await ready()`);
  const sn = starryNightCache;
  const scope = sn.flagToScope(code.language);

  if (typeof scope !== 'string') {
    throw new Error(`language ${code.language} not found`);
  }

  const parsed = sn.highlight(raw, scope);
  return parsed.children.map(child => colorRecurse(child, (options === null || options === void 0 ? void 0 : options.codeStyle) ?? {})).flat().map(({
    color,
    ...rest
  }) => ({ ...rest,
    color: color === '' ? '#c9d1d9' : color
  }));
}

const rules = new Map(Object.entries(_darkStyle.default).map(([k, v]) => [k, new Map(Object.entries(v))]));
const styleMap = [[['pl-s'], s => {
  var _s$stringContent;

  return (_s$stringContent = s.stringContent) === null || _s$stringContent === void 0 ? void 0 : _s$stringContent.text;
}], [['pl-s', 'pl-pds'], s => {
  var _s$stringPunctuation;

  return (_s$stringPunctuation = s.stringPunctuation) === null || _s$stringPunctuation === void 0 ? void 0 : _s$stringPunctuation.text;
}], [['pl-c'], s => {
  var _s$comment;

  return (_s$comment = s.comment) === null || _s$comment === void 0 ? void 0 : _s$comment.text;
}], [['pl-smi'], s => {
  var _s$variable;

  return (_s$variable = s.variable) === null || _s$variable === void 0 ? void 0 : _s$variable.text;
}], [['pl-v'], s => {
  var _s$parameter;

  return (_s$parameter = s.parameter) === null || _s$parameter === void 0 ? void 0 : _s$parameter.text;
}], [['pl-s', 'pl-sr'], s => {
  var _s$regexp, _s$regexp2;

  return ((_s$regexp = s.regexp) === null || _s$regexp === void 0 ? void 0 : _s$regexp.content) ?? ((_s$regexp2 = s.regexp) === null || _s$regexp2 === void 0 ? void 0 : _s$regexp2.text);
}], [['pl-s', 'pl-sr', 'pl-pds'], s => {
  var _s$regexp3, _s$regexp4;

  return ((_s$regexp3 = s.regexp) === null || _s$regexp3 === void 0 ? void 0 : _s$regexp3.brackets) ?? ((_s$regexp4 = s.regexp) === null || _s$regexp4 === void 0 ? void 0 : _s$regexp4.text);
}], [['pl-s', 'pl-sr', 'pl-k'], s => {
  var _s$regexp5, _s$regexp6;

  return ((_s$regexp5 = s.regexp) === null || _s$regexp5 === void 0 ? void 0 : _s$regexp5.flags) ?? ((_s$regexp6 = s.regexp) === null || _s$regexp6 === void 0 ? void 0 : _s$regexp6.text);
}], [['pl-c1'], s => {
  var _s$literal;

  return (_s$literal = s.literal) === null || _s$literal === void 0 ? void 0 : _s$literal.text;
}], [['pl-k'], s => {
  var _s$keyword;

  return (_s$keyword = s.keyword) === null || _s$keyword === void 0 ? void 0 : _s$keyword.text;
}], [['pl-en'], s => {
  var _s$entityName;

  return (_s$entityName = s.entityName) === null || _s$entityName === void 0 ? void 0 : _s$entityName.text;
}], [['pl-ent'], s => {
  var _s$entityNameTag;

  return (_s$entityNameTag = s.entityNameTag) === null || _s$entityNameTag === void 0 ? void 0 : _s$entityNameTag.text;
}]];

function findAssociated(styleMap, target, notFound) {
  let best = notFound;
  let specificity = 0;

  for (const [keys, value] of styleMap) {
    if (keys.length <= specificity) {
      continue;
    }

    let keyIndex = 0;
    let targetIndex = 0;

    while (targetIndex < target.length && keyIndex < keys.length) {
      const key = keys[keyIndex];
      const targetKey = target[targetIndex];

      if (targetKey === key) {
        keyIndex++;
      }

      targetIndex++;
    }

    if (keyIndex === keys.length) {
      // all keys found
      best = value;
      specificity = keys.length;
    }
  }

  return best;
}

function getColor(classList, codeStyle) {
  const fn = findAssociated(styleMap, classList, () => undefined);
  const result = fn(codeStyle);

  if (result != null) {
    return result;
  }

  const styles = rules.get(`.${classList.at(-1)}`);
  console.assert(((styles === null || styles === void 0 ? void 0 : styles.size) ?? 0) <= 1, `more styles than just color`);
  return styles === null || styles === void 0 ? void 0 : styles.get('color');
}

function colorRecurse(parsed, codeStyle, classList = []) {
  if (parsed.type === 'text') {
    var _codeStyle$base;

    return [{
      code: parsed.value,
      color: ((_codeStyle$base = codeStyle.base) === null || _codeStyle$base === void 0 ? void 0 : _codeStyle$base.text) ?? ''
    }];
  } else if (parsed.type === 'element') {
    var _parsed$properties;

    const className = ((_parsed$properties = parsed.properties) === null || _parsed$properties === void 0 ? void 0 : _parsed$properties.className) ?? [];
    const color = getColor(classList.concat(className), codeStyle);
    const children = parsed.children.map(child => colorRecurse(child, codeStyle, classList.concat(className)));
    const result = [];

    const emit = () => {
      if (temp !== '') {
        if (color != '') {
          result.push({
            code: temp,
            color
          });
        } else {
          result.push({
            code: temp,
            color: ''
          });
        }

        temp = '';
      }
    };

    let temp = '';

    for (const child of children) {
      for (const item of child) {
        if (item.color === '') {
          temp += item.code;
        } else {
          emit();
          result.push(item);
        }
      }

      emit();
    }

    emit();
    return result;
  } else {
    throw new Error();
  }
}

function toString(tokens) {
  return tokens.map(token => token.code).join('');
}

function reindent(code, indent = 0) {
  var _code$spans$at;

  if (typeof code === 'string') {
    if (code.at(0) !== '\n') {
      return ' '.repeat(indent) + code;
    }

    const regex = /^ +/gm;
    const matches = code.matchAll(regex);
    let min = Infinity;

    for (const match of matches) min = Math.min(min, match[0].length);

    const undent = new RegExp(`^ {${min}}`, 'gm');
    const result = code.substring(1).replace(undent, ' '.repeat(indent));
    return result;
  }

  if (((_code$spans$at = code.spans.at(0)) === null || _code$spans$at === void 0 ? void 0 : _code$spans$at.at(0)) !== '\n') return code;
  const regex = /\n */g;
  const matches = code.spans.flatMap(span => [...span.matchAll(regex)]);
  let min = Infinity;

  for (const [match] of matches) min = Math.min(min, match.length - 1);

  const undentRegex = new RegExp(`\n {${min}}`, 'g');
  const spans = code.spans.map(span => span.replace(undentRegex, '\n' + ' '.repeat(indent)));
  spans[0] = spans[0].substring(1);
  let index = 0;
  const nodes = [];

  for (const node of code.nodes) {
    const before = spans[index];
    const preindentRegex = /\n *$/;
    const indentation = before.match(preindentRegex);

    if (indentation != null) {
      spans[index] = spans[index].replace(preindentRegex, '\n');
      nodes.push(reindent(node, indentation[0].length - 1));
    } else {
      nodes.push(node);
    }

    index++;
  }

  return {
    language: code.language,
    spans,
    nodes
  };
}

function integrate(code) {
  if (typeof code === 'string') return code;
  return String.raw({
    raw: code.spans
  }, ...code.nodes.map(n => integrate(n)));
}

function isLevelEquivilant(one, two) {
  if (one === null || two === null) return false;

  if (typeof one === 'string' && typeof two !== 'string' || typeof one !== 'string' && typeof two === 'string') {
    return false;
  }

  if (typeof one === 'string' && typeof two === 'string') {
    return one === two;
  }

  if (typeof one !== 'string' && typeof two !== 'string') {
    if (one.spans.length !== two.spans.length) return false;
    return one.spans.every((span, i) => span === two.spans[i]);
  }

  throw new Error('Could not determine equivilance of nodes');
}

function chars(tokens) {
  return tokens.flatMap(({
    code,
    color
  }) => {
    return code.split('').map(c => ({
      code: c,
      color
    }));
  });
}

function tokens(chars) {
  const result = [];
  let token = null;

  for (const char of chars) {
    if (token == null) {
      token = char;
    } else {
      if (char.code === '\n') {
        result.push(token);
        token = null;
      } else if (token.color === char.color && token.morph === char.morph) {
        token.code += char.code;
      } else {
        result.push(token);
        token = char;
      }
    }
  }

  if (token != null) result.push(token);
  return result;
}

function diff(start, end, options) {
  start = reindent(start);
  end = reindent(end);
  const startParsed = chars(parse(start, options));
  const endParsed = chars(parse(end, options));
  let index = 0;
  let endex = 0;
  const result = [];

  function recurse(one, two) {
    const progress = (l, r) => {
      if (r == null) r = l;
      const startIndex = index;

      while (index < startIndex + l.length && index < startParsed.length) {
        result.push({
          code: startParsed[index].code,
          color: startParsed[index].color,
          morph: l === r ? 'retain' : 'delete'
        });
        index++;
      }

      const endIndex = endex;

      while (endex < endIndex + r.length && two !== '' && endex < endParsed.length) {
        if (r !== l) {
          result.push({
            code: endParsed[endex].code,
            color: endParsed[endex].color,
            morph: 'create'
          });
        }

        endex++;
      }
    };

    if (isLevelEquivilant(one, two)) {
      if (one == null || two == null) {
        throw new Error('equivilant nodes should not be null');
      }

      if (typeof one === 'string') {
        progress(one);
      } else {
        if (typeof two === 'string') throw new Error();

        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n]);
          recurse(one.nodes[n], two.nodes[n]);
        }

        progress(one.spans.at(-1) ?? '');
      }
    } else {
      if (typeof one === 'string' && typeof two === 'string') {
        progress(one, two);
      } else if (typeof one === 'string') {
        progress(one, '');
      } else if (typeof two === 'string') {
        progress('', two);
      }

      if (one != null && typeof one !== 'string') {
        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n], '');
          recurse(one.nodes[n], null);
        }

        progress(one.spans.at(-1) ?? '', '');
      }

      if (two != null && typeof two !== 'string') {
        for (let n = 0; n < two.nodes.length; n++) {
          progress('', two.spans[n]);
          recurse(null, two.nodes[n]);
        }

        progress('', two.spans.at(-1));
      }
    }
  }

  recurse(start, end);
  let [sat, sln, eat, eln] = [0, 0, 0, 0];
  const positioned = result.map(({
    code,
    color,
    morph
  }) => {
    const value = {
      code,
      color,
      morph,
      from: morph === 'create' ? null : [sat, sln],
      to: morph === 'delete' ? null : [eat, eln]
    };

    if (code === '\n') {
      if (morph !== 'create') {
        sat = 0;
        sln++;
      }

      if (morph !== 'delete') {
        eat = 0;
        eln++;
      }
    } else {
      if (morph !== 'create') {
        sat += (0, _wcwidth.default)(code);
      }

      if (morph !== 'delete') {
        eat += (0, _wcwidth.default)(code);
      }
    }

    return value;
  });
  const tokenized = tokens(positioned);
  return tokenized.filter(({
    code
  }) => {
    return code.length !== 0 && !/^ +$/.test(code);
  });
}